#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/matrix_sparse.hpp>
#include <boost/numeric/ublas/io.hpp>
#include <boost/numeric/ublas/operation.hpp>
#include <boost/numeric/ublas/storage.hpp>
#include <boost/numeric/ublas/matrix_proxy.hpp>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <cassert>
#include <queue>

#include<arpack++/arssym.h>

typedef float value_t;

value_t threshold=0.5; //sensibility of the clusterer
value_t tol=1.e-4; //tol for the eigensolver
value_t zerotol=1.e-3; //entries smaller than this are considered to be 0

using namespace boost::numeric::ublas;
using std::cerr;
using std::cout;
using std::endl;

std::vector<std::string> labels;

typedef compressed_matrix<value_t,row_major> backend_t;

class mymatrix_t:public backend_t{
public:
  mymatrix_t(size_t rows, size_t cols, size_t nnz=0):
    backend_t(rows,cols,nnz){};

  void multvet(value_t *v, value_t *w){ //performs Mv, stores result in w. Needed by arpack++
    //actually, v and multvet() should be const, but arpack++ demands this exact signature. Poor design?
    shallow_array_adaptor<value_t> v_adap(size1(), v);
    vector<value_t, shallow_array_adaptor<value_t> > v_vec(size1(), v_adap);
    shallow_array_adaptor<value_t> w_adap(size1(), w);
    vector<value_t, shallow_array_adaptor<value_t> > w_vec(size1(), w_adap);
    axpy_prod(*this, v_vec, w_vec, true);  // w = A * v
  };
  
  bool is_symmetric() const{ //ineffective, used only for testing
    for(backend_t::const_iterator1 it1=begin1();it1!=end1();++it1)
      for(backend_t::const_iterator2 it2=it1.begin();it2!=it1.end();++it2)
	if((*this)(it2.index1(),it2.index2())!=(*this)(it2.index2(),it2.index1())) return false;
    return true;
  };

  std::vector<size_t> orig_pos; //original indices of the elements of the matrix -- preserved during the permutations
};

void eigsplit(std::queue<mymatrix_t*> &q);
void labeller(const mymatrix_t &m, ARSymStdEig<value_t,mymatrix_t> &eig); //labels clusters
void labeller_single(size_t orig_pos); //labels singletons
void labeller_double(size_t orig_pos1, size_t orig_pos2); //labels doubletons

inline int sign(int x){
  return (x>0?1:(x<0?-1:0));
}

int main(int argc, char **argv){
  //reads graph
  int rows, cols, nnz;
  std::ifstream f(argv[1]);
  f>>rows>>cols>>nnz;
  if(rows!=cols){cerr<<"hmm... rectangular matrix -- or wrong parsing?\n";return 1;}

  mymatrix_t *m=new mymatrix_t(rows,cols,nnz); //XXX: use shared_ptr?
  cerr<<"Reading matrix..."<<endl;
  //for(size_t i=0;i<nnz;++i){
  while(f.good()){
    size_t ii,jj;value_t dd;
    f>>ii>>jj>>dd;
    if(dd==0) continue; //skips zero entries
    (*m)(ii,jj)=dd;
  }
  f.close();

  cerr<<"Reading labels..."<<endl;
  //reads labels
  std::ifstream labelsfile(argv[2]);
  labels.reserve(rows);
  char buf[256];
  for(int i=0;i<rows;++i){
    labelsfile.getline(buf,256);
    labels.push_back(buf);
  }
  labelsfile.close();

  //test print
  //matrix_slice<mymatrix_t> smallm(m,slice(0,1,3),slice(0,1,3));
  //cout<<smallm<<endl;
  //XX: to be changed for nonsymmetric matrices

  //cout<<*m<<endl;

  //correctly sets orig_pos
  m->orig_pos.reserve(rows);
  for(size_t i=0;i<rows;++i)
    m->orig_pos.push_back(i);

  cout<<"Testing multvet speed. If this does not end soon, forget about getting the clusters... ";
  value_t *v=new value_t[rows],*w=new value_t[rows];
  for(size_t i=0;i<rows;++i)
    v[i]=1;
  m->multvet(v,w);
  cout<<"done"<<endl;

  std::queue<mymatrix_t*> q;
  q.push(m);
  while(!q.empty()) eigsplit(q);
  
  return 0;
}

void eigsplit(std::queue<mymatrix_t*> &q){  //pops from the queue, splits, pushes back the two new matrices
  mymatrix_t *m=q.front();
  size_t n=m->size1(); //for shortness
  if(n<=2){
    if(n==2){
      value_t same_block=(*m)(0,0)+(*m)(1,1), cross_block=(*m)(0,1)+(*m)(1,0);
      if(cross_block<threshold*same_block){ //split further
	labeller_single(m->orig_pos[0]);
	labeller_single(m->orig_pos[1]);
      }
      else labeller_double(m->orig_pos[0],m->orig_pos[1]);
    }
    else if(n==1)
      labeller_single(m->orig_pos[0]);
    else if(n==0)
      cerr<<"hmm... matrix of size 0?"<<endl;
    delete m;
    q.pop();
    return;
  }
  ARSymStdEig<value_t,mymatrix_t> eig(n,2,m,&mymatrix_t::multvet,tol); //XXX:DBG
  //cerr<<"Starting eigenproblem with n="<<n<<endl;
  eig.FindEigenvectors();
  assert(eig.ConvergedEigenvalues()==2);
  //cout<<eig.Eigenvalue(1)<<";"<<eig.Eigenvalue(0)<<endl;

  //finds an eigenvector not totally positive / negative
  //int smaller=eig.Eigenvalue(1)>eig.Eigenvalue(0) ? 0 : 1; //SPD case, all eigvals should be positive
  int smaller;
  bool has_pos, has_neg, has_zero;
  {
    size_t i;
    for(i=0;i<eig.ConvergedEigenvalues();++i){
      has_pos=has_neg=has_zero=false;
      for(int j=0;j<n;++j){
	if(eig.Eigenvector(i,j)>zerotol) has_pos=true;
	else if(eig.Eigenvector(i,j)<-zerotol) has_neg=true;
	else has_zero=true;
      }
      if((has_pos && has_neg) || (has_pos && has_zero) || (has_zero && has_neg)) break;
    }
    if(i==2){
      cout<<"Eigenvecs:"<<endl;
      for(int i=0;i<eig.ConvergedEigenvalues();++i){
	cout<<endl<<"Eigenvalue: "<<eig.Eigenvalue(i)<<endl;
	for(int j=0;j<eig.GetN();++j)
	  cout<<eig.Eigenvector(i,j)<<endl;
      }
      exit(1);
    }
    smaller=i;
  }
   
  std::vector<int> p;p.reserve(n);
  //will fill p[i] = \pm (new index of [i] +1) in the new submatrix

  int ip=1,im=-1;
  if(has_pos){
    for(size_t j=0;j<n;++j)
      p.push_back(eig.Eigenvector(smaller,j)>zerotol?ip++:im--);
  } else {
    for(size_t j=0;j<n;++j)
      p.push_back(eig.Eigenvector(smaller,j)<-zerotol?ip++:im--);
  }
  
  //creates the two new submatrices
  mymatrix_t *mp,*mm;
  mp=new mymatrix_t(ip-1,ip-1);
  mm=new mymatrix_t(-im-1,-im-1);

  //#if 0 //DBG
  //XXX: check what happens when size=0
  if(ip-1==0 || -im-1==0){
    cout<<"Eigenvecs:"<<endl;
    for(int i=0;i<eig.ConvergedEigenvalues();++i){
      cout<<endl<<"Eigenvalue: "<<eig.Eigenvalue(i)<<endl;
      for(int j=0;j<eig.GetN();++j)
	cout<<eig.Eigenvector(i,j)<<endl;
	}
  }
  //#endif

  //fills in orig_pos
  mp->orig_pos.reserve(ip-1);
  mm->orig_pos.reserve(-im-1);
  {
    std::vector<size_t>::iterator posit;
    std::vector<int>::iterator it;
    for(it=p.begin(),posit=m->orig_pos.begin();
	it!=p.end();++it,++posit)
      (*it>0?mp:mm)->orig_pos.push_back(*posit);
  }
  //fills in the new matrices
  //in the meantime, evaluates the same-block and cross-block norms, for the stopping criterion
  value_t same_block=0,cross_block=0;
  
  for(mymatrix_t::iterator1 it1=m->begin1();it1!=m->end1();++it1)
    for(mymatrix_t::iterator2 it2=it1.begin();it2!=it1.end();++it2)
      if(sign(p[it2.index2()])==sign(p[it2.index1()])){
	int i=abs(p[it2.index1()])-1,j=abs(p[it2.index2()])-1;
	(p[it2.index1()]>0?mp:mm)->insert_element(i,j,*it2);
	same_block+=*it2;
      }
      else{
	cross_block+=*it2;
      }
  if(cross_block<threshold*same_block){ //cross-block part is negligible, the splitting is meaningful
    delete m;
    q.pop(); //removes m
    q.push(mp);
    q.push(mm);
  }
  else{
    labeller(*m,eig); //XXX 1-smaller=larger
    delete m;
    q.pop();
  }
}

void labeller(const mymatrix_t &m, ARSymStdEig<value_t,mymatrix_t> &eig){
  cout<<"****Cluster****"<<endl;
  for(int i=0;i!=m.size1();++i)
    cout<<labels[m.orig_pos[i]]<<endl;
}
void labeller_single(size_t orig_pos){
  cout<<"****Cluster****"<<endl;
  cout<<labels[orig_pos]<<endl;
} 
void labeller_double(size_t orig_pos1, size_t orig_pos2){
  cout<<"****Cluster****"<<endl;
  cout<<labels[orig_pos1]<<endl;
  cout<<labels[orig_pos2]<<endl;
}
